---
title: "OT"
author: "Jean Feydy"
institute: ""
topic: ""
theme: "metropolis"
colortheme:
fonttheme:
fontsize: 11pt
linkcolor: black
urlcolor: blue
linkstyle: bold
aspectratio: 169
titlegraphic: 
logo:
date:
section-titles: false
toc: false
---



## Recent works

`\vspace*{.5cm}`{=latex}

::: columns
:::: {.column width=35%}

`\begin{center}`{=latex}

![](images/lungs.png){height=4.7cm}

Lung **registration**.
`\end{center}`{=latex}

::::

:::: {.column width=64%}

`\begin{center}`{=latex}

![](images/brain_arterial_network.png){height=4.7cm}

Interventional **radiology**.

`\end{center}`{=latex}

::::

:::

`\vspace*{.3cm}\begin{center}`{=latex}

$\Longrightarrow~~$ **Accessible** to you guys, but $\BR{barely anyone else}$.

`\end{center}`{=latex}




## Recent works 

`\vspace*{.5cm}`{=latex}

::: columns

:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/fracture.png){height=4.7cm}

Orthopedic **surgery**.
`\end{center}`{=latex}

::::

:::: {.column width=50%}

`\begin{center}`{=latex}

![](images/survie_france.png){height=4.7cm}

**Public** health.
`\end{center}`{=latex}

::::


:::

`\vspace*{.3cm}\begin{center}`{=latex}

$\Longrightarrow~~$ **Accessible** to you guys, but $\BR{barely anyone else}$.

`\end{center}`{=latex}





## Recent works 

`\vspace*{.5cm}`{=latex}

::: columns
:::: {.column width=35%}

`\begin{center}`{=latex}

![](images/APD_3.png){height=4.7cm}

**Metallurgy**.
`\end{center}`{=latex}

::::

:::: {.column width=64%}

`\begin{center}`{=latex}

![](images/cells.png){height=4.7cm}

Swarms of incompressible **cells**.

`\end{center}`{=latex}

::::

:::

`\vspace*{.3cm}\begin{center}`{=latex}

$\Longrightarrow~~$ **Accessible** to you guys, but $\BR{barely anyone else}$.

`\end{center}`{=latex}



## HeKA : a translational research team for public health

`\begin{center}`{=latex}

![](images/HeKA.pdf){height=7cm}

`\end{center}`{=latex}



## Geometric data analysts are in a delicate position

Our constraints:

1. Differential geometry is **not** part of the $\BR{mainstream curriculum}$ \
   $\Longrightarrow~~$ High **entry cost** for students and users. \
    \

2. **Credibility** $\Longleftrightarrow$ $\BR{Performance}$ and high-resolution figures \
   $\Longrightarrow~~$ Constant work to **keep up** with new technology. \
    \

3. We are already $\BR{very busy}$ \
   $\Longrightarrow~~$ Our **career incentives** do **not** reward long-term software **maintenance**.


## Today’s talk -- some practical answers to three pressing questions

1. Which **language** and **libraries** should I use?

2. Is my code still going to run in **2030**?

3. How do I get $\BR{rewarded}$ for all of that extra work?


## Which language should I use?

 \
The $\BR{C++ era}$ (2000-2015):

- High-performance C++ was **necessary** to handle $\BR{3D data}$.
- **Monolithic** code-bases with a lot of $\BR{inertia}$, cryptic to scientists.
- The Visualization ToolKit, the Computational Geometry Algorithms Library...

 \
The $\BR{Python era}$ (since 2015):

- **Modular** and **inter-operable** tools via dictionaries and Nu-mPy arrays.
- **Permissive** open source licences create trust.
- Scikit-learn, Scikit-image, PyVista, Vedo...

## Which language should I use?

Domain-specific languages are fine too:

- **R** is data-centric: native idiom for biologists and medical doctors.
- **Julia** is convenient for numerical analysis.

 \
But **Python** is the $\BR{lingua franca}$ for **gluing** pipelines together:

1. Identify the key **building blocks** in your method.
2. Implement them in the language that suits you best.
3. Write a $\BR{Python interface}$ -- now super easy.

`\vspace*{.5cm}\begin{center}`{=latex}

$\Longrightarrow~~$ Speak **French**, **German** or **Hindi** at home... but publish in $\BR{English}$.

`\end{center}`{=latex}

  
## The KeOps library: efficient support for symbolic matrices, with Joan and Benjamin


::: columns

:::: {.column width=70%}

`\solution{KeOps}`{=latex} -- `\textcolor{blue}{\texttt{www.kernel-operations.io}}`{=latex}: `\vspace*{.2cm}`{=latex}

- For PyTorch, NumPy, Matlab and R, on **CPU and GPU**.
- **Automatic differentiation**.
- Just-in-time **compilation** of **optimized** C++ schemes, \
  triggered for every new **reduction**: sum, min, etc.

\
If the formula "F" is simple ($\leqslant$ 100 arithmetic operations): \
$\qquad$ "100k $\times$ 100k" computation  $\rightarrow$  10ms -- 100ms, \
$\qquad$     "1M $\times$ 1M"    computation  $\rightarrow\phantom{\text{0m}}$ 1s -- 10s.

`\vspace*{.5cm}`{=latex}

`\begin{center}`{=latex}

Hardware ceiling of $\text{10}^{\text{12}}$ operations/s. \
**$\times \textbf{10}$ to $\times \textbf{100}$ speed-up** vs standard GPU implementations \
for a wide range of problems.

`\end{center}`{=latex}

::::

:::: {.column width=25%}

`\begin{center}`{=latex}



```{=latex}
\begin{tikzpicture}[scale=.9]
	% Left border:
	\draw[draw=black, fill=red!20] (-.45,0) rectangle ++(.3,3);
	\draw[draw=black, fill=red!80] (-.45,9*.15) rectangle ++(.3,.15);
	\draw[step=.15, color=gray] (-.45,0) grid ++(.3,3);
	\draw[draw=black] (-.45,0) rectangle ++(.3,3);
	%
	% Top border:
	\draw[draw=black, fill=green!20] (0,3.15) rectangle ++(3,.3);
	\draw[draw=black, fill=green!80] (6*.15,3.15) rectangle ++(.15,.3);
	\draw[step=.15, color=gray] (0,3.15) grid ++(3,.3);
	\draw[draw=green!50!black] (0,3.15) rectangle ++(3,.3);
	%
	\draw[draw=black,thick] (0,0) rectangle ++(3,3);
	%
	\draw[->, thick, color=red!80!black] (.5*.15, 9.5*.15) -- (5*.15, 9.5*.15);
	\draw[->, thick, color=green!50!black] (6.5*.15, 19.5*.15) -- (6.5*.15, 11*.15);
	%
	\draw[fill=blue!40] (6*.15,9*.15) rectangle ++(.15,.15);
	\node[inner sep=0pt, anchor = mid, color=blue!50!black] at (1.95,1.4) {\normalsize{F(\,\textcolor{red!80!black}{$\text{x}_\text{i}$}\,,\,\textcolor{green!50!black}{$\text{y}_\text{j}$}\,)}};
\end{tikzpicture}
```

**Symbolic matrix** \
Formula + data
 

`\end{center}`{=latex}
`\vspace*{.5cm}`{=latex}


- Distances `d(\textcolor{red!80!black}{$\text{x}_\text{i}$},\textcolor{green!50!black}{$\text{y}_\text{j}$})`{=latex}.
- Kernel `k(\textcolor{red!80!black}{$\text{x}_\text{i}$},\textcolor{green!50!black}{$\text{y}_\text{j}$})`{=latex}.
- Numerous transforms.


::::

:::


## Yet another python compiler?

Many impressive tools out there (Numba, Triton, Halide, Taichi...):

- Focus on **generality** (software + hardware).
- Increasingly easy to use via e.g. PyTorch 2.0.


KeOps fills a $\BR{scientific niche}$ (like FFT libraries):

- Focus on a **single major bottleneck**: geometric interactions.
- **Agnostic** with respect to Euclidean / non-Euclidean formulas.
- Fully compatible with PyTorch, NumPy, R.
- Can actually be $\BR{used by mathematicians}$ (**600k+** downloads).

`\begin{center}`{=latex}

KeOps is a **bridge** between geometers (with a maths background) \
and compiler experts (with a CS background).

`\end{center}`{=latex}



## Which libraries should I use?

Exciting libraries get $\BR{killed}$ all the time **:-(**

**Theano** (2008-2017):

  - $\BR{Pioneering}$ deep learning library: Python + Autodiff + GPU.
  - Created and maintained in Montreal (MILA).
  - Development stopped when **PyTorch** became available.

**Taichi** (2017-2023):

  - $\BR{Awesome}$ Python dialect for 3D shape processing and graphics, 25k GitHub stars.
  - **PhD thesis** of Yuanming Hu at MIT, now **CEO** of Meshy.
  - Active development stopped last summer.


## Tip \#1: look at the developers' long term incentives

 \
**PyTorch** (Meta) -- sending all the $\SIM{right signals}$:

- Business strategy on AI is to make it an **open source commodity**.
- $\BR{Transparent governance structure}$, PyTorch foundation.
- **Extensive** internal documentation.


 \
**JAX and TensorFlow** (Google) -- several $\BR{red flags}$:

- Business strategy on AI is to protect the **Google search monopoly** and **GCP**.
- $\BR{Opaque}$ governance structure, $~\texttt{killedbygoogle.com}$.
- **Minimal** internal documentation.



## Tip \#2: Implement a future-proof interface

**Insulate users** from deprecations:

- Numpy arrays.
- Human-readable files.

$\BR{User-centric}$ design:

- Principle of **least surprise**.
- Write $\BR{tutorials}$ -- a feature that is not documented **does not exist**.
- Plain, descriptive names:


```{=latex}
\vspace*{-.5cm}\begin{center}

\begin{tabular}{ r c l }
 Kernel & $\longrightarrow$ & \small{\texttt{\textbf{covariance}}} \\ 
 Splines & $\longrightarrow$ & \small{\texttt{\textbf{deformation}(covariance="thin plate spline")}} \\  
 \textbf{LDDMM} & $\longrightarrow$ & \small{\texttt{deformation(covariance="gaussian", scale=2, \textbf{n\_steps=10})}}    
\end{tabular}

\end{center}
```



## Problem: software rots


::: columns
:::: {.column width=58%}

`\vspace*{1.7cm}`{=latex}
Some personal nightmares: \
 \


- CMake, **Boost**...
- Nvidia **actively deprecates** "old" GPUs.
- $\texttt{torch.solve(A, B)} = B^{-1}A \rightarrow A^{-1}B$.


`\begin{center}`{=latex}

 \
Without $\BR{constant gardening}$, \
software breaks after 3-5 years.

`\end{center}`{=latex}


::::

:::: {.column width=40%}

`\begin{center}`{=latex}

![](images/maison_abandonnee.jpg){height=6.5cm}

*Some of my old GitHub repositories.*

`\end{center}`{=latex}

::::

:::




## Research you're proud of should be in a library



::: columns
:::: {.column width=58%}

 \
 \
**\#1** -- Include your model in a **pre-existing** library:

- $\BR{Outsource maintenance}$, gain visibility.
- Permissive licenses are key: MIT, BSD...


 \
**\#2** -- Develop and maintain **your own** library:

- Be realistic: focus on your **core expertise**. 
- Bet on $\BR{interoperability}$ with other packages.
- Freedom for you, minimize risk for users.

`\begin{center}`{=latex}


 \
$\Longrightarrow$ Agree on a $\BR{consistent interface}$ with \
the community and **keep your word**. 

`\end{center}`{=latex}

::::

:::: {.column width=40%}

`\begin{center}`{=latex}

![](images/bibliotheque_paris.jpg){height=6.5cm}

*A **professional** storage facility.*

`\end{center}`{=latex}

::::

:::

## GeomLoss: scaling up optimal transport to anatomical data

```{=latex}

\begin{center} 
	Progresses of the last decade add up to a 
	$\times \textbf{100}$ - $\times \textbf{1000}$ acceleration:
	
	Sinkhorn GPU
	$\xrightarrow{\times \text{10}}$
	+ KeOps
	$\xrightarrow{\times \text{10}}$
	+ Annealing
	$\xrightarrow{\times \text{10}}$
	+ Multi-scale
	
	
	With a precision of 1\%, on a modern gaming GPU:\\~
	\end{center}

	\vspace*{-.25cm}
	\begin{tabular}{cccc}
	\begin{minipage}{.2\linewidth}
	\begin{center}
	~\\
	\texttt{pip install geomloss}\\
	+\\
	gaming GPU\\ (1 000 €)
	\end{center}
	\end{minipage}
	&
	\begin{minipage}{.1\linewidth}
	~\\
	$\Longrightarrow$
	\end{minipage}
	&
	\begin{minipage}{.3\linewidth}
	\includegraphics[height=3.5cm]{images/dragon_ball/lowres.png}
	\end{minipage}
	&
	\begin{minipage}{.3\linewidth}
	\includegraphics[height=3.5cm]{images/dragon_ball/highres.png}
	\end{minipage}
	\\
	&&
	10k points in 30-50ms
	&
	\textbf{100k points in 100-200ms}
	\end{tabular}
```

## GeomLoss: going forward

Current landscape in computational optimal transport:

- **Python Optimal Transport (POT)**: tons of tutorials, but slow solvers from 2015.
- Mérigot, Lévy, De Goes: super-fast OT solvers for **physics**. 
- Schmitzer, GeomLoss: super-fast OT solvers for **geometric data**.
- Massive $\BR{waste of time for newcomers}$ in the field.

How to solve the issue:

- Agree on a $\BR{common interface}$.
- Include GeomLoss and others as **optional backends** in POT.
- **Automated benchmark** website to highlight "solved" and "open" problems.

$\Longrightarrow~$ Put egos aside, move forward as a community. \
$\longrightarrow~$ Only possible because we are not judged by our h-index.


## But Jean... I'm just a mathematician!



::: columns
:::: {.column width=58%}

 \
 \
Writing good code is easy now! \
Use $\BR{professional tools}$: \
 \

- Black and Ruff **beautify** your code.
- Pytest and Hypothesis **find bugs**.
- Copilot **writes documentation**.
- Sphinx creates a **clean website**.
- GitHub actions **deploy automagically**.

 \
Check out $~\BR{\texttt{scientific-python.org}}$.

::::

:::: {.column width=40%}

`\begin{center}`{=latex}

![](images/power_tools.jpg){height=6.5cm}

*Invest in **power** tools.*

`\end{center}`{=latex}

::::

:::



## But Jean... I'm just a mathematician!

 \
$\BR{Why should I bother?}$

- If you don't code your method first, **no one will**.
- Get to meet a wide range of **exciting users**.
- Open up **career paths** for students.

 \
$\BR{Publish or perish?}$

- French open source software awards from the Ministry of research.
- At INRIA, **clear incentives** for software development.
- Career paths for **research engineers** in academia?



## From Deformetrica to scikit-shapes (with an 's')



::: columns
:::: {.column width=58%}

`\vspace*{.5cm}`{=latex}
$\BR{scikit-shapes:}$
 \

- Follows the tips above!
- Named after **scikit-image**: a reference library for classical image processing.
- **Abstracts** multiscaling and feature extraction.
- **Foundations** are now solid (Louis Pujol).
- Funded by INRIA and Prairie.
 
 \
**Next steps:**

- LDDMM and elastic metrics.
- GPMM and functional maps.
- **Research** on robustness and modularity.


::::

:::: {.column width=40%}

`\begin{center}`{=latex}

![](images/scikit_shapes.png){height=6.5cm}

$\BR{\texttt{scikit-shapes.github.io}}$ \
Check it out in 2025!

`\end{center}`{=latex}

::::

:::






## Our community is judged by its software output


::: columns

:::: {.column width=48%}

`\begin{center}`{=latex}

![](images/tower_of_babel.jpg){height=6.5cm}

*The C++ tower of Babel.*

`\end{center}`{=latex}

::::
:::: {.column width=48%}

`\begin{center}`{=latex}

![](images/asterix_market.png){height=6.5cm}

*The Python market of ideas.*

`\end{center}`{=latex}

::::

:::


`\vspace*{.3cm}\begin{center}`{=latex}

Major challenge: beyond goodwill, create $\textbf{sustainable \BR{open} business models}$. \
Are **universities** hostile environments? Kitware (VTK), Tutte Institute (UMAP), INRIA... 

`\end{center}`{=latex}


## Documentation and tutorials are available online

```{=latex}
\begin{center}
$\Longrightarrow$\qquad \url{shape-analysis.github.io}\qquad $\Longleftarrow$
\vspace*{.15cm}

\includegraphics[width=.75\linewidth,frame]{images/shape_analysis.png}

%\vspace*{.15cm}

Monthly seminar, videos on YouTube.
\end{center}

```



# References


